#pragma warning( disable : 4312 )

#include <allegro.h>
#include <winalleg.h>
#include <iostream>
#include <vector>
#include <list>

using namespace std;

char *levelNames[5] = {"gfx/level1.bmp", "gfx/level2.bmp", "gfx/level3.bmp", "gfx/level4.bmp", "gfx/level5.bmp"};
const int maxLevels = 5;

#include "GameState.h"


GameState::GameState()
{
	allegro_init();
	set_color_depth(16);
	set_gfx_mode(GFX_AUTODETECT,640,480,0,0);

	install_timer();
	install_keyboard();
	install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, NULL);

	currentLevel = 0;
	score = 0;
	timer = 5*60*100;

	// load sounds
	laserSound = load_wav("sfx/laser.wav");			laserSoundPlaying = false;
	splatSound = load_wav("sfx/splat.wav");			splatSoundPlaying = false;
	shakeSound = load_wav("sfx/shakeexplode.wav");	shakeSoundPlaying = false;
	naniteSound1 = load_wav("sfx/nanites.wav");
	oofSound	= load_wav("sfx/oof.wav");

	bgMusic = load_midi("sfx/wickedsystem.mid");

	// load game graphics
	gameGraphics = load_bitmap("gfx/swarmgraph.bmp",NULL);

	// create pre_rendered level bitmap
	preRenderedLevel = create_bitmap(3200,3200);

	// create the double buffer
	doubleBuffer = create_bitmap(640,480);

}

GameState::~GameState()
{
	destroy_sample(laserSound);
	destroy_sample(splatSound);
	destroy_sample(shakeSound);
	destroy_sample(naniteSound1);
	destroy_midi(bgMusic);

	remove_sound();
	remove_keyboard();
	remove_timer();

	// unload game graphics
	destroy_bitmap(gameGraphics);
	destroy_bitmap(doubleBuffer);
	destroy_bitmap(preRenderedLevel);
}

bool GameState::RenderTitle()
{
	clear(doubleBuffer);

	BITMAP *titleGraphic = load_bitmap("gfx/title.bmp",NULL);

	draw_sprite(doubleBuffer,titleGraphic,128,180);
	draw_sprite(screen,doubleBuffer,0,0);

	clear_keybuf();
	while(!keypressed()) yield_timeslice();

	FadeOut();

	if( (readkey() >> 8) == KEY_ESC)
		return false;
	else
		return true;

	destroy_bitmap(titleGraphic);
}

void GameState::RenderStory()
{
	clear(doubleBuffer);

	BITMAP *storyGraphic = load_bitmap("gfx/story.bmp",NULL);

	draw_sprite(doubleBuffer,storyGraphic,128,180);
	draw_sprite(screen,doubleBuffer,0,0);

	clear_keybuf();
	while(!keypressed()) yield_timeslice();

	FadeOut();

	destroy_bitmap(storyGraphic);

}


void GameState::RenderEnding()
{
	char tempString[256];

	clear(doubleBuffer);
	BITMAP *endingGraphic = load_bitmap("gfx/ending.bmp",NULL);

	draw_sprite(doubleBuffer, endingGraphic, 128,180);
	
	sprintf(tempString,"Your Final Score: %d",score);
	textout_ex(doubleBuffer,font,tempString,160,300,makecol(139,142,206),-1);


	draw_sprite(screen,doubleBuffer,0,0);

	clear_keybuf();
	while(!keypressed()) yield_timeslice();

	FadeOut();

	destroy_bitmap(endingGraphic);
}

void GameState::RenderHero()
{
	int sourceX=0;
	int sourceY=0;


	if(hero.actionState == hit)
	{
		switch(hero.frame) {
			case 0: sourceX = 4; sourceY = 0; break;
			case 1: sourceX = 4; sourceY = 1; break;
		}
	}
	else if(hero.actionState == dead)
	{
		sourceX = 5;
		sourceY = 2;
	}
	else if(hero.actionState == dying)
	{
		switch(hero.frame) {
						case 0: sourceX = 2; sourceY = 0; break;
						case 1: sourceX = 4; sourceY = 1; break;
						case 2: sourceX = 4; sourceY = 0; break;
						case 3: sourceX = 4; sourceY = 0; break;
						case 4: sourceX = 5; sourceY = 0; break;
						case 5: sourceX = 5; sourceY = 1; break;
						case 6: sourceX = 5; sourceY = 2; break;
			}
	}
	else
	switch(hero.directionalState)
	{
	case up:
		switch(hero.frame) {
							case 0: sourceX = 0;
									sourceY = 0;
									break;
							case 1: sourceX = 0;
									sourceY = 1;
									break;
							case 2: sourceX = 0;
									sourceY = 0;
									break;
							case 3: sourceX = 0;
									sourceY = 2;
									break;
							}
		break;

	case down:
				switch(hero.frame) {
							case 0: sourceX = 2;
									sourceY = 0;
									break;
							case 1: sourceX = 2;
									sourceY = 1;
									break;
							case 2: sourceX = 2;
									sourceY = 0;
									break;
							case 3: sourceX = 2;
									sourceY = 2;
									break;
							}
		break;

	case left:
				switch(hero.frame) {
							case 0: sourceX = 3;
									sourceY = 0;
									break;
							case 1: sourceX = 3;
									sourceY = 1;
									break;
							case 2: sourceX = 3;
									sourceY = 0;
									break;
							case 3: sourceX = 3;
									sourceY = 2;
									break;
							}
		break;

	case right:
				switch(hero.frame) {
							case 0: sourceX = 1;
									sourceY = 0;
									break;
							case 1: sourceX = 1;
									sourceY = 1;
									break;
							case 2: sourceX = 1;
									sourceY = 0;
									break;
							case 3: sourceX = 1;
									sourceY = 2;
									break;
							}
		break;
	}

	sourceX *= 32;
	sourceY *= 32;

	if(hero.invincibleCount % 2 == 0 || hero.invincibleCount <= 0)
		{
		masked_blit(gameGraphics,doubleBuffer,sourceX,sourceY,hero.x-16-cameraX,hero.y-16-cameraY,32,32);
		}

	if(hero.invincibleCount > 0)
		hero.invincibleCount--;
}

void GameState::RenderNanites()
{
	int sourceX = 0;
	int sourceY = 0;

	for (unsigned int i = 0; i < nanites.size(); i++)
	{
		switch(nanites[i].directionalState)
		{
		case up:		sourceX = 6; break;
		case right:		sourceX = 8; break;
		case left:		sourceX = 7; break;
		case down:		sourceX = 9; break;
		}
		
		sourceX *= 32;
		sourceY = (nanites[i].frame+(currentLevel-1)*2) * 32;
		
		masked_blit(gameGraphics,doubleBuffer,sourceX,sourceY,nanites[i].x - 16-cameraX, nanites[i].y - 16-cameraY,32,32);
	}
}

void GameState::RenderBeam()
{
	static bool flicker = true;

	drawing_mode(DRAW_MODE_TRANS,NULL,0,0);

	if(flicker && hero.actionState == shooting)
		switch(hero.directionalState)
			{
			case left:
				rectfill(doubleBuffer,0,hero.y-6-cameraY, hero.x-12-cameraX, hero.y+6-cameraY, makecol(0,0,255));
				break;

			case right:
				rectfill(doubleBuffer,hero.x-cameraX+6,hero.y-6-cameraY,639,hero.y+6-cameraY,makecol(0,0,255));
				break;
		
			case up:
				rectfill(doubleBuffer,hero.x-6-cameraX,0,hero.x+6-cameraX,hero.y-16-cameraY,makecol(0,0,255));
				break;

			case down:
				rectfill(doubleBuffer,hero.x-6-cameraX,hero.y-cameraY,hero.x+6-cameraX,479,makecol(0,0,255));
				break;

			default:
				break;
			}

	drawing_mode(DRAW_MODE_SOLID,NULL,0,0);

	//flicker = !flicker;
}


void GameState::RenderLevel()
{

	// render starting at the camera a chunk of 640 x 480
	blit(preRenderedLevel,doubleBuffer,cameraX,cameraY,0,0,640,480);

}


void GameState::PlayMusic()
{
	play_midi(bgMusic,1);
}

void GameState::StopMusic()
{
	stop_midi();
}

void GameState::NextLevel()
{
  LoadLevel(levelNames[currentLevel]);
  currentLevel++;
}

void GameState::LoadLevel(char *filename)
{
	levelMap = load_bitmap(filename,NULL);
	clear(preRenderedLevel);

	CreateHero();
	nanites.clear();
	
	// process special colors
	for(int j = 0; j < levelMap->h; j++)
	for(int i = 0; i < levelMap->w; i++)
	{
		if(getpixel(levelMap,i,j) == makecol(0,0,255))
		{
			hero.x = i * 32+16;
			hero.y = j * 32+16;

			putpixel(levelMap,i,j,makecol(255,255,255));
			blit(gameGraphics,preRenderedLevel,4*32,(3+currentLevel)*32,i*32,j*32,32,32);
		}
		else if(getpixel(levelMap,i,j) == makecol(255,0,0))
		{
			Nanite newNanite;
			newNanite.directionalState = DirectionalState(rand()%4);
			newNanite.disabled = false;
			newNanite.dx = rand()%3-1;
			newNanite.dy = rand()%3-1;
			newNanite.frame = 0;
			newNanite.x = i*32+16;
			newNanite.y = j*32+16;
			nanites.push_back(newNanite);
			putpixel(levelMap,i,j,makecol(255,255,255));
			blit(gameGraphics,preRenderedLevel,4*32,(3+currentLevel)*32,i*32,j*32,32,32);
		}
		else
		{
			if(getpixel(levelMap,i,j) == makecol(0,0,0) || getpixel(levelMap,i,j) == makecol(255,255,0))
				{
				//rectfill(preRenderedLevel,i*32,j*32,i*32+31,j*32+31,makecol(128,128,128));
				blit(gameGraphics,preRenderedLevel,rand()%3*32,(3+currentLevel)*32,i*32,j*32,32,32);
				}
			else if(getpixel(levelMap,i,j) == makecol(0,255,0))
				blit(gameGraphics,preRenderedLevel,3*32,(3+currentLevel)*32,i*32,j*32,32,32);
			else
				blit(gameGraphics,preRenderedLevel,4*32,(3+currentLevel)*32,i*32,j*32,32,32);
		}

	}
	
	// center the camera around the player
	cameraX = hero.x-320;
	cameraY = hero.y-200;
	levelCompleted = false;
	gameOver = false;

}


void GameState::PerformCollisionMods()
{
	int p;

	// make sure the hero doesn't walk into a wall

	if( getpixel(levelMap,hero.x/32,hero.y/32) == makecol(0,0,0)
		|| getpixel(levelMap,hero.x/32,hero.y/32) == makecol(255,255,0))
	{
		switch(hero.directionalState)
		{
		case up:	hero.y++; break;
		case down:	hero.y--; break;
		case left:	hero.x++; break;
		case right: hero.x--; break;
		}
	}
	else if(getpixel(levelMap,hero.x/32,hero.y/32) == makecol(0,255,0))
	{
	  levelCompleted = true;	
	}

	// make sure the nanites don't walk through walls and can cause premature hero death :-)
	// also check to see if they suffer death by mega beam

	for(int i = 0; i < nanites.size(); i++)
	{

	if(hero.actionState == shooting)
	switch(hero.directionalState)
	{
	case right:
		if(nanites[i].x > hero.x && nanites[i].y > hero.y - 6 && nanites[i].y < hero.y + 6)
			{
			for(p = 0; p < 30; p++) {BloodParticle newP; newP.life = 30; newP.dx = rand()%3-1; newP.dy = rand()%3-1; newP.x = nanites[i].x, newP.y = nanites[i].y; particles.push_back(newP);}
			play_sample(splatSound, 64, 0, 1000, 0); 
			nanites.erase(nanites.begin() + i);
			score += 125 * currentLevel;
			}
			break;
	
	case left:
		if(nanites[i].x < hero.x && nanites[i].y > hero.y - 6 && nanites[i].y < hero.y + 6)
			{
			for(p = 0; p < 30; p++) {BloodParticle newP; newP.life = 30; newP.dx = rand()%3-1; newP.dy = rand()%3-1; newP.x = nanites[i].x, newP.y = nanites[i].y;particles.push_back(newP);}
			play_sample(splatSound, 64, 0, 1000, 0); 
			nanites.erase(nanites.begin() + i);
			score += 125 * currentLevel;
			}
			break;

	case up:
		if(nanites[i].y < hero.y && nanites[i].x > hero.x - 6 && nanites[i].x < hero.x + 6)
			{
			for(p = 0; p < 30; p++) {BloodParticle newP; newP.life = 30; newP.dx = rand()%3-1; newP.dy = rand()%3-1; newP.x = nanites[i].x, newP.y = nanites[i].y;particles.push_back(newP);}
			play_sample(splatSound, 64, 0, 1000, 0); 
			nanites.erase(nanites.begin() + i);
			score += 125 * currentLevel;
			}
			break;

	case down:
		if(nanites[i].y > hero.y && nanites[i].x > hero.x - 6 && nanites[i].x < hero.x + 6)
			{
			for(p = 0; p < 30; p++) {BloodParticle newP; newP.life = 30; newP.dx = rand()%3-1; newP.dy = rand()%3-1; newP.x = nanites[i].x, newP.y = nanites[i].y;particles.push_back(newP);}
			play_sample(splatSound, 64, 0, 1000, 0); 
			nanites.erase(nanites.begin() + i);
			score += 125 * currentLevel;
			}
			break;

	}

	if( abs(nanites[i].x - hero.x) < 16 && abs(nanites[i].y - hero.y) < 16 && hero.actionState != dead && hero.invincibleCount <= 0)
		{
			hero.lifePower-=10;
			if(hero.lifePower <= 0)
				hero.actionState = dying;
			else
			{
				hero.invincibleCount = 100;
				hero.actionState = hit;
			}
		}

	if( getpixel(levelMap,nanites[i].x/32, nanites[i].y/32) == makecol(0,0,0) )
		switch(nanites[i].directionalState)
		{
		case up:	nanites[i].y++; nanites[i].directionalState=down; break;
		case down:	nanites[i].y--; nanites[i].directionalState=up; break;
		case left:	nanites[i].x++; nanites[i].directionalState=right; break;
		case right:	nanites[i].x--; nanites[i].directionalState=left; break;
		}
	}
}



void GameState::CreateHero()
{
	hero.lifePower = 100;
	hero.beamPower = 100;
	hero.directionalState = down;
	hero.movementState = still;
	hero.actionState = donothing;
	hero.invincibleCount = 100;
	hero.frame = 0;
	hero.x = 320;
	hero.y = 200;
}


void GameState::NextHeroState()
{

	if(hero.actionState == hit)
	{
		if(hero.frame == 2)
			{
			hero.actionState = donothing;
			hero.frame = 0;
			}
	}
	if(hero.actionState != dead && hero.actionState != dying)
	{
	if(key[KEY_SPACE])		
		{

		if(hero.beamPower >= 0)
			hero.actionState = shooting;
		else
			hero.actionState = donothing;
		
		hero.beamPower-=1;
		}
	else
		{
		hero.actionState = donothing;
		if(hero.beamPower < 100) hero.beamPower++;
		}

	if(key[KEY_LEFT])			hero.directionalState = left, hero.movementState = moving;
	else if(key[KEY_RIGHT])		hero.directionalState = right, hero.movementState = moving;
	else if(key[KEY_UP])		hero.directionalState = up, hero.movementState = moving;
	else if(key[KEY_DOWN])		hero.directionalState = down, hero.movementState = moving;
	else hero.movementState = still;
	}
	else if(hero.actionState == dying)
	{
	  if(hero.frame == 5)
		  hero.actionState = dead;
		  hero.movementState = still;
	}


	if(laserSoundPlaying == false && hero.actionState == shooting)
	{
		laserSoundPlaying = true;
		play_sample(laserSound, 255, 0, 1000, 1);
	}
	else
	if(laserSoundPlaying == true && hero.actionState != shooting)
	{
		laserSoundPlaying = false;
		stop_sample(laserSound);
	}
}

void GameState::NextNaniteState()
{
	for(int i = 0; i < nanites.size(); i++)
	{
		if(rand()%200 == 0)
			nanites[i].directionalState = DirectionalState(rand()%4);
	}
}

void GameState::NextParticleState()
{
	for(int i = 0; i < particles.size(); i++)
	{
		particles[i].x += particles[i].dx;
		particles[i].y += particles[i].dy;
		particles[i].life--;
	
		if(particles[i].life <= 0)
			particles.erase(particles.begin()+i);
	}
}


bool GameState::NextState()
{
	static int counter = 0;

	NextHeroState();
	NextNaniteState();
	NextParticleState();

	if(counter % 30 == 0)
	{
	NextHeroFrame();
	NextNaniteFrame();
	}

	if(counter % 1 == 0)
	{
	NextHeroMove();
	NextNaniteMove();
	
	PerformCollisionMods();
	}

	// at random, play the nanite sound
	if(rand()%500 == 0) play_sample(naniteSound1, 64,0,1000,0);

	yield_timeslice();	// let windows do something
	rest(1), counter++;
	
	timer--;
	if(timer <= 0)
		gameOver = true;

	if(key[KEY_ESC])
		gameOver = true;

	if(gameOver || levelCompleted)
		return false;	// to quit
	else
		return true;	// to continue
}	


void GameState::NextHeroMove()
{
	if(hero.movementState == moving && hero.actionState != dead && hero.actionState != dying)
	switch(hero.directionalState)
	{
	case up:	hero.y--; break;
	case left:	hero.x--; break;
	case right: hero.x++; break;
	case down:	hero.y++; break;
	}
}


void GameState::NextNaniteFrame()
{
	for(unsigned int i = 0; i < nanites.size(); i++)
	{
		nanites[i].frame = (nanites[i].frame + 1) % 2;
	}
}



void GameState::NextHeroFrame()
{
	if(hero.actionState == hit)
	{
		if(hero.frame < 2) hero.frame++;
	}
	else if(hero.actionState == dying)
	{	
		if(hero.frame < 6) hero.frame++;
	}
	else if(hero.actionState == dead)
			hero.frame = 0;
	else if(hero.movementState == moving)
	{
	if(hero.directionalState == left)
		hero.frame = (hero.frame + 1) % 4;
	else if(hero.directionalState == right)
		hero.frame = (hero.frame + 1) % 4;
	else if(hero.directionalState == up)
		hero.frame = (hero.frame + 1) % 4;
	else if(hero.directionalState == down)
		hero.frame = (hero.frame + 1) % 4;
	}
	else
	{
		hero.frame = 0;
	}
}


void GameState::NextNaniteMove()
{
	for(int i = 0; i < nanites.size(); i++)
		switch(nanites[i].directionalState)
		{
		case up:		nanites[i].y--; break;
		case right:		nanites[i].x++; break;
		case left:		nanites[i].x--; break;
		case down:		nanites[i].y++; break;
		}
}


void GameState::FadeOut()
{

	for(int i = 0; i < 2024; i++)
	{
		int randomY = rand()%640;
		line(screen,randomY,0,randomY,479,makecol(0,0,0));
		yield_timeslice();
		rest(1);
	}

}


void GameState::RenderState()
{
	//clear_to_color(doubleBuffer,makecol(0,0,0));

	// manage camera to center around player but only if everything
	// can be shown

	cameraX = hero.x - 320;
	if(cameraX < 0) cameraX = 0;
	if(cameraX > 3200-640) cameraX = 3200-640;

	cameraY = hero.y - 200;
	if(cameraY < 0) cameraY = 0;
	if(cameraY > 3200 - 480) cameraY = 3200 - 480;


	RenderLevel();
	RenderHero();
	RenderNanites();
	RenderBeam();
	RenderParticles();
	RenderHud();

	if(timer < 10000)
	{
		blit(doubleBuffer,screen,0,0,rand()%5,rand()%5,635,475);
		if(shakeSoundPlaying == false) 
			{
			play_sample(shakeSound,64,0,1000,1);
			shakeSoundPlaying = true;
			}
	}
	else
		blit(doubleBuffer,screen,0,0,0,0,640,480);
}

void GameState::RenderParticles()
{
	drawing_mode(DRAW_MODE_TRANS,NULL,0,0);

	for(int i = 0; i < particles.size(); i++)
	{
		circlefill(doubleBuffer, particles[i].x-cameraX, particles[i].y-cameraY, 3,makecol(0,0,255));
	}

	drawing_mode(DRAW_MODE_SOLID,NULL,0,0);
}

void GameState::RenderHud()
{
	  char tempString[256];

	  set_trans_blender(128, 128, 128, 128);
      draw_trans_sprite(doubleBuffer, levelMap, 640-101, 0);

	  if(hero.lifePower > 0)
		rectfill(doubleBuffer,640-101,100,hero.lifePower+640-101,110,makecol(255,0,0));

	  if(hero.beamPower > 0)
	    rectfill(doubleBuffer,640-101,111,hero.beamPower+640-101,120,makecol(0,0,255));

	  sprintf(tempString,"Score: %d",score);
	  textout_ex(doubleBuffer, font, tempString, 640-101,125, makecol(255,255,255), -1);

	  sprintf(tempString,"Timer: %d",timer/100);

	  if(timer < 10000)
		textout_ex(doubleBuffer, font, tempString, 640-101,140, makecol(255,0,0), -1);
	  else
	  	textout_ex(doubleBuffer, font, tempString, 640-101,140, makecol(255,255,255), -1);


	  circlefill(doubleBuffer,640-101+hero.x/32,hero.y/32,2,makecol(255,255,255));
}


int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    GameState *gameState = NULL;
	
	gameState = new GameState();
	
	if(gameState->RenderTitle())
	{
	gameState->RenderStory();
	// perform menu stuff
	do
	{
		// load first level
		gameState->NextLevel();
		gameState->PlayMusic();
		
		while(gameState->NextState())
		{
			if(key[KEY_P])
				save_bitmap("screenshot.bmp",screen,NULL);

			gameState->RenderState();
		}

		gameState->FadeOut();
		gameState->StopMusic();
		
		if(gameState->currentLevel == 5)
		{
			// show end of game graphics
			gameState->RenderEnding();
			gameState->gameOver = true;
		}

	} while(!gameState->gameOver);
	}


	delete gameState;

return 0;
}